from datetime import datetime, timezone
import json
import logging
import logging.handlers
import queue

from pydantic import BaseModel


class JsonFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        try:
            if hasattr(record, "event") and isinstance(record.event, BaseModel):
                return record.event.model_dump_json()

            if hasattr(record, "event"):  # Event is a dictionary
                return json.dumps(record.event, ensure_ascii=False)

        except Exception as e:
            return self.serialization_error_log(record, e)

        return json.dumps(
            {
                "timestamp": datetime.now(tz=timezone.utc).isoformat(),
                "level": record.levelname,
                "message": record.getMessage(),
            }
        )

    def serialization_error_log(self, record: logging.LogRecord, e: Exception) -> str:
        return json.dumps(
            {
                "code": "LOGGER_JSON_SERIALIZATION_ERROR",
                "timestamp": datetime.now(tz=timezone.utc).isoformat(),
                "level": record.levelname,
                "message": f"Failed to serialize event to JSON: {str(e)}",
            }
        )

def create_json_logger() -> logging.Logger:
    """
    Creates and configures a JSON logger.

    Returns:
        logging.Logger: Configured JSON logger.
    """

    json_event_logger = logging.getLogger("json_logger")

    if not json_event_logger.handlers:
        json_event_logger.setLevel(logging.INFO)
        log_queue: queue.Queue[logging.LogRecord] = queue.Queue(-1)  # infinite size
        queue_handler = logging.handlers.QueueHandler(log_queue)

        stream_handler = logging.StreamHandler()
        listener = logging.handlers.QueueListener(log_queue, stream_handler)
        stream_handler.setFormatter(JsonFormatter())
        json_event_logger.addHandler(queue_handler)
        json_event_logger.propagate = False

        listener.start()
        json_event_logger.listener = listener  # type: ignore[attr-defined]

    return json_event_logger



import re
from typing import Any, Optional, TypeVar
from uuid import uuid4

from httpx import Headers


def build_url(*url_components: Optional[str]) -> str:
    """
    Builds a URL from the provided components, removing any unintended extra slashes and ignoring
    components that are None.
    """
    compacted_url_components = [
        component for component in url_components if component is not None
    ]
    components_joined_by_slash = "/".join(compacted_url_components)

    url = _remove_double_slashes_not_preceded_by_columns(components_joined_by_slash)

    return url


def _remove_double_slashes_not_preceded_by_columns(url: str) -> str:
    if url.startswith("https:"):
        return re.sub(r"(?<!https:)/+", "/", url)
    elif url.startswith("http:"):
        return re.sub(r"(?<!http:)/+", "/", url)


T = TypeVar("T")


def get_all_missing_keys_in_dict(dict_: dict[T, Any], keys: list[T]) -> list[T]:
    """Returns a list of keys that are missing from the provided dictionary."""
    missing_keys = [key for key in keys if key not in dict_]
    return missing_keys

def mask_str(string: str, keep: int = 4) -> str:
    """
    Mask all but the last `keep` characters of a str.
    Example: abcdefghijkl -> ********ijkl
    """
    if len(string) <= keep:
        return "*" * len(string)
    masked = "*" * (len(string) - keep) + string[-keep:]
    return masked

def create_short_unique_id(length: int = 12) -> str:
    return uuid4().hex[:length]

def obfuscate_headers(headers: Headers | dict[str, str]) -> dict[str, str]:
    """
    Return a copy of headers with sensitive values partially obfuscated.
    Sensitive fields are detected by common HTTP header naming conventions.
    """
    obfuscated = {}

    for key, value in headers.items():
        lower_key = key.lower()

        is_sensitive_header = (
            any(  # Check if the header key contains any of the sensitive keywords
                sensitive in lower_key
                for sensitive in [
                    "authorization",
                    "api-key",
                    "apikey",
                    "secret",
                    "token",
                    "password",
                    "passwd",
                ]
            )
        )

        if is_sensitive_header:
            obfuscated[key] = mask_str(value)

        else:
            obfuscated[key] = value

    return obfuscated




from abc import ABC, abstractmethod
from typing import Any, Optional

class AbstractLocalCache(ABC):

    @abstractmethod
    def get(self, key: str) -> Optional[Any]:
        raise NotImplementedError

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        raise NotImplementedError


from enum import StrEnum
from typing import Any, Callable, Coroutine, Mapping, Optional

from httpx import URL, AsyncBaseTransport, AsyncClient, Request, Response
from httpx._client import EventHook
from httpx._config import (
    DEFAULT_LIMITS,
    DEFAULT_MAX_REDIRECTS,
    DEFAULT_TIMEOUT_CONFIG,
    Limits,
)
from httpx._types import (
    AuthTypes,
    CertTypes,
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    TimeoutTypes,
)

from core_infinity_stones.core.utils.helpers import obfuscate_headers
from core_infinity_stones.logging import (
    IncomingResponseEvent,
    Logger,
    OutgoingRequestEvent,
)


class LoggingVerbosity(StrEnum):
    NONE = "NONE"
    """Do not log anything."""

    LOW = "LOW"
    """Only log request and response metadata (method, URL, status code, duration)."""

    MEDIUM = "MEDIUM"
    """Log request and response headers in addition to metadata."""

    HIGH = "HIGH"
    """Log request and response bodies truncated to a 1000 chars in addition to metadata and headers."""

    FULL = "FULL"
    """Log request and response full bodies in addition to metadata and headers."""

    @staticmethod
    def form_string(string: str) -> "LoggingVerbosity":
        try:
            return LoggingVerbosity[string.upper()]
        except KeyError:
            return LoggingVerbosity.NONE

    @property
    def should_log_headers(self) -> bool:
        return self in {LoggingVerbosity.MEDIUM, LoggingVerbosity.HIGH}

    @property
    def should_log_body(self) -> bool:
        return self in {LoggingVerbosity.HIGH, LoggingVerbosity.FULL}

    @property
    def should_truncate_body(self) -> bool:
        return self != LoggingVerbosity.FULL


class HttpxClientWithLogging(AsyncClient):
    def __init__(
        self,
        logger: Logger,
        auth: AuthTypes | None = None,
        params: QueryParamTypes | None = None,
        headers: HeaderTypes | None = None,
        cookies: CookieTypes | None = None,
        cert: CertTypes | None = None,
        http1: bool = True,
        http2: bool = False,
        timeout: TimeoutTypes = DEFAULT_TIMEOUT_CONFIG,
        follow_redirects: bool = False,
        limits: Limits = DEFAULT_LIMITS,
        max_redirects: int = DEFAULT_MAX_REDIRECTS,
        event_hooks: None | (Mapping[str, list[EventHook]]) = None,
        base_url: URL | str = "",
        transport: AsyncBaseTransport | None = None,
        trust_env: bool = True,
        default_encoding: str | Callable[[bytes], str] = "utf-8",
        request_verbosity_level: LoggingVerbosity = LoggingVerbosity.HIGH,
        response_verbosity_level: LoggingVerbosity = LoggingVerbosity.HIGH,
        path_to_request_verbosity_level_map: Optional[Mapping[str, LoggingVerbosity]] = None,
        path_to_response_verbosity_level_map: Optional[Mapping[str, LoggingVerbosity]] = None,
    ) -> None:
        event_hooks_with_logging: Optional[Mapping[str, list[Callable[..., Any]]]] = None

        request_event_hooks = event_hooks.get("request", []) if event_hooks else []
        response_event_hooks = event_hooks.get("response", []) if event_hooks else []

        request_logger = self.create_request_logger(
            request_verbosity_level, path_to_request_verbosity_level_map, logger
        )

        response_logger = self.create_response_logger(
            response_verbosity_level, path_to_response_verbosity_level_map, logger
        )

        event_hooks_with_logging = {
            "request": [request_logger, *request_event_hooks],
            "response": [response_logger, *response_event_hooks],
        }

        super().__init__(
            auth=auth,
            params=params,
            headers=headers,
            cookies=cookies,
            cert=cert,
            http1=http1,
            http2=http2,
            timeout=timeout,
            follow_redirects=follow_redirects,
            limits=limits,
            max_redirects=max_redirects,
            event_hooks=event_hooks_with_logging or event_hooks,
            base_url=base_url,
            transport=transport,
            trust_env=trust_env,
            default_encoding=default_encoding,
        )

    def create_request_logger(
        self,
        default_verbosity_level: LoggingVerbosity,
        path_to_verbosity_level_map: Optional[Mapping[str, LoggingVerbosity]],
        logger: Logger,
    ) -> Callable[[Request], Coroutine[Any, Any, None]]:
        async def request_logger(request: Request) -> None:
            verbosity = ( 
                path_to_verbosity_level_map.get(request.url.path, default_verbosity_level)
                if path_to_verbosity_level_map
                else default_verbosity_level
            )

            if verbosity == LoggingVerbosity.NONE:
                return

            request_body = None

            if verbosity.should_log_body:
                request_body = (
                    request.content[:1000] + bytes("... [truncated]", "utf-8")
                    if verbosity.should_truncate_body
                    else request.content
                )

            headers = (
                obfuscate_headers(request.headers)
                if verbosity.should_log_headers
                else None
            )

            logger.info(
                OutgoingRequestEvent(
                    method=request.method,
                    url=str(request.url),
                    headers=headers,
                    body=request_body,
                    verbosity_level=verbosity,
                )
            )

        return request_logger

    def create_response_logger(
        self,
        default_verbosity_level: LoggingVerbosity,
        path_to_verbosity_level_map: Optional[Mapping[str, LoggingVerbosity]],
        logger: Logger,
    ) -> Callable[[Response], Coroutine[Any, Any, None]]:

        async def response_logger(response: Response) -> None:
            verbosity = (
                path_to_verbosity_level_map.get(response.request.url.path, default_verbosity_level)
                if path_to_verbosity_level_map
                else default_verbosity_level
            )

            if verbosity == LoggingVerbosity.NONE:
                return

            await response.aread()  # Ensure the response content is read
            request = response.request
            response_body = None

            if verbosity.should_log_body:
                response_body = (
                    response.text[:1000]
                    if verbosity.should_truncate_body
                    else response.text
                )

            headers = (
                obfuscate_headers(response.headers)
                if verbosity.should_log_headers
                else None
            )

            logger.info(
                IncomingResponseEvent(
                    method=request.method,
                    url=str(request.url),
                    status_code=response.status_code,
                    duration_ms=response.elapsed.total_seconds() * 1000,
                    headers=headers,
                    body=response_body,
                    verbosity_level=verbosity,
                )
            )

        return response_logger



from datetime import datetime, timezone
from enum import StrEnum
from typing import Any, Optional
from uuid import UUID

from pydantic import BaseModel, Field

from core_infinity_stones.core.utils.helpers import create_short_unique_id
from core_infinity_stones.errors.base_error import LocalizedMessage, Severity


class RequestDetails(BaseModel):
    path: Optional[str] = None
    method: Optional[str] = None
    query_params: Optional[dict[str, str]] = None
    user_agent: Optional[str] = None

class LoggerContext(BaseModel):
    trace_id: str
    request_details: "RequestDetails"
    context_metadata: dict[str, Any]
    topics: set[str]

class EventLevel(StrEnum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


class Event(BaseModel):
    id: str = Field(default_factory=lambda: create_short_unique_id())
    code: str
    message: Optional[str] = None
    details: Optional[dict[str, Any]] = None
    topic: Optional[str] = None


class EventWithTracesDetails(BaseModel):
    id: str
    trace_id: str
    level: EventLevel
    path: Optional[str] = None
    method: Optional[str] = None
    query_params: Optional[dict[str, Any]] = None
    user_agent: Optional[str] = None
    code: str
    topics: Optional[set[str]] = None
    message: Optional[str] = None
    context_metadata: Optional[dict[str, Any]] = None
    details: Optional[dict[str, Any]] = None
    sampling_percentage: int
    timestamp: str

    @classmethod
    def from_event(
        cls,
        event: Event,
        trace_id: str,
        level: EventLevel,
        sampling_percentage: int,
        context_metadata: dict[str, Any],
        topics: set[str],
        request_details: Optional[RequestDetails] = None,
    ) -> "EventWithTracesDetails":
        event_topics = topics.copy()

        if event.topic:
            event_topics.add(event.topic)

        return EventWithTracesDetails(
            id=event.id,
            trace_id=trace_id,
            level=level,
            path=request_details.path if request_details else None,
            method=request_details.method if request_details else None,
            query_params=request_details.query_params if request_details else None,
            user_agent=request_details.user_agent if request_details else None,
            code=event.code,
            topics=event_topics if event_topics else None,
            message=event.message,
            context_metadata=context_metadata if context_metadata else None,
            details=event.details,
            sampling_percentage=sampling_percentage,
            timestamp=datetime.now(tz=timezone.utc).isoformat(),
        )


class ErrorEvent(BaseModel):
    id: str
    trace_id: str
    level: EventLevel
    path: Optional[str] = None
    method: Optional[str] = None
    query_params: Optional[dict[str, Any]] = None
    user_agent: Optional[str] = None
    severity: Severity
    code: str
    topics: Optional[set[str]] = None
    message: str
    context_metadata: Optional[dict[str, Any]] = None
    details: Optional[dict[str, Any]] = None
    occurred_while: Optional[str] = None
    caused_by: Optional[dict[str, Any]] = None
    status_code: int
    public_code: str
    public_message: LocalizedMessage
    public_details: Optional[dict[str, Any]] = None
    sampling_percentage: int
    timestamp: str


# Common Events

class IncomingRequestEvent(Event):
    def __init__(
        self,
        method: str,
        url: str,
        headers: Optional[dict[str, str]] = None,
        body: Optional[Any] = None,
    ) -> None:
        details = {
            "method": method,
            "url": url,
            "headers": headers,
            "body": body,
        }

        super().__init__(code="INCOMING_REQUEST", details=details)


class OutgoingRequestEvent(Event):
    def __init__(
        self,
        verbosity_level: str,
        method: str,
        url: str,
        headers: Optional[dict[str, str]] = None,
        body: Optional[Any] = None,
    ) -> None:
        details = {
            "method": method,
            "url": url,
            "headers": headers,
            "body": body,
            "verbosity_level": verbosity_level,
        }

        super().__init__(code="OUTGOING_REQUEST", details=details)


class OutgoingResponseEvent(Event):
    def __init__(
        self,
        method: str,
        url: str,
        status_code: int,
        duration_ms: float,
        headers: Optional[dict[str, str]] = None,
        body: Optional[Any] = None,
    ) -> None:
        details = {
            "method": method,
            "url": url,
            "status_code": status_code,
            "duration_ms": duration_ms,
            "headers": headers,
            "body": body,
        }

        super().__init__(code="OUTGOING_RESPONSE", details=details)


class IncomingResponseEvent(Event):
    def __init__(
        self,
        verbosity_level: str,
        method: str,
        url: str,
        status_code: int,
        duration_ms: float,
        headers: Optional[dict[str, str]] = None,
        body: Optional[Any] = None,
    ) -> None:
        details = {
            "method": method,
            "url": url,
            "status_code": status_code,
            "duration_ms": duration_ms,
            "headers": headers,
            "body": body,
            "verbosity_level": verbosity_level,
        }

        super().__init__(code="INCOMING_RESPONSE", details=details)


from typing import Generic, TypeVar
from uuid import UUID

from pydantic import BaseModel

from core_infinity_stones.networking.response_handler import HttpxResponse

T = TypeVar("T", bound=BaseModel)


class BaseModelResponseDecoder(Generic[T]):
    def __init__(self, model: type[T]):
        self.model = model

    def decode_as_object(self, response: HttpxResponse) -> T:
        return self.model(**response.json())

    def decode_as_list(self, response: HttpxResponse) -> list[T]:
        return [self.model(**element) for element in response.json()]

    def decode_as_str_to_model_map(self, response: HttpxResponse) -> dict[str, T]:
        return {key: self.model(**value) for key, value in response.json().items()}

    def decode_as_uuid_to_model_map(self, response: HttpxResponse) -> dict[UUID, T]:
        return {
            UUID(key): self.model(**value) for key, value in response.json().items()
        }



# setup.py
from setuptools import find_packages, setup

long_description = ""
# Read long description from README.md
with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="core_infinity_stones",
    version="0.1.15",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={"core_infinity_stones": ["py.typed"]},
    python_requires=">=3.6",
    install_requires=[
        "annotated-types==0.7.0",
        "anyio==4.8.0",
        "certifi==2024.12.14",
        "h11==0.14.0",
        "httpcore==1.0.8",
        "httpx==0.28.1",
        "idna==3.10",
        "pydantic==2.10.6",
        "pydantic_core==2.27.2",
        "sniffio==1.3.1",
        "typing_extensions==4.12.2",
    ],
)



from typing import Optional
from core_infinity_stones.errors.base_error import HttpError, LocalizedMessage, Severity


class ResponseStatusCodeError(HttpError):
    def __init__(
        self,
        url: str,
        status_code: int,
        debug_description: str,
        messages_by_status_codes: Optional[dict[int, str]] = None,
    ):
        message = (
            messages_by_status_codes.get(status_code, "")
            if messages_by_status_codes and status_code in messages_by_status_codes
            else "An error occurred while processing the request."
        )

        super().__init__(
            status_code=status_code,
            severity=Severity.MEDIUM if 400 <= status_code < 500 else Severity.HIGH,
            code="RESPONSE_STATUS_CODE_ERROR",
            message=LocalizedMessage(en=message, ar=message),
            debug_message=debug_description,
            debug_details={"url": url, "status_code": status_code},
        )



import random
import traceback
from contextvars import ContextVar
from datetime import datetime, timezone
from typing import Any, Optional

from core_infinity_stones.errors.base_error import HttpError
from core_infinity_stones.logging.native_logger import create_json_logger
from core_infinity_stones.logging.schemas import (
    ErrorEvent,
    Event,
    EventLevel,
    EventWithTracesDetails,
    LoggerContext,
    RequestDetails,
)

logger_context: ContextVar[Optional[LoggerContext]] = ContextVar("logger_context", default=None)

class Logger:
    def __init__(
        self,
        event_codes_to_sampling_percentages_map: Optional[dict[str, int]] = None,
    ):
        """
        Initialize the Logger instance.

        Args:
            trace_id (str): The trace ID for the current request or context.

            logger (NativeLogger): The logger instance to use for logging events.

            request_details (Optional[RequestDetails]): Details about the request,
              such as path, method, and query parameters.

            context_metadata (dict[str, Any]): A dictionary to hold additional context metadata
                that will be included in every log entry.

            topics (set[str]): A set of topics associated with the logger instance.

            event_codes_to_sampling_percentages_map (Optional[dict[str, int]]): A mapping of event
                codes to their sampling percentages.
                This is a dictionary where keys are event codes and values are integers between 0 and 100,
                representing the probability of logging the specified event.
                If None, all events will be logged with 100% sampling.
                If an event code is not present in the map, it will default to 100% sampling.
        """
        self.logger = create_json_logger()
        self.event_codes_to_sampling_percentages_map = (
            event_codes_to_sampling_percentages_map
        )

    def set_logger_context(
        self,
        trace_id: str,
        request_details: RequestDetails,
        context_metadata: Optional[dict[str, Any]] = None,
        topics: Optional[set[str]] = None,
    ) -> None:
        logger_context.set(
            LoggerContext(
                trace_id=trace_id,
                request_details=request_details,
                context_metadata=context_metadata or {},
                topics=topics or set(),
            )
        )

    @property
    def trace_id(self) -> str:
        ctx = logger_context.get()

        if ctx is None:
            return ""

        return ctx.trace_id or ""

    @property
    def request_details(self) -> RequestDetails:
        ctx = logger_context.get()

        if ctx is None:
            return RequestDetails()

        return ctx.request_details

    @property
    def context_metadata(self) -> dict[str, Any]:
        ctx = logger_context.get()

        if ctx is None:
            return {}
        return ctx.context_metadata or {}

    @property
    def topics(self) -> set[str]:
        ctx = logger_context.get()

        if ctx is None:
            return set()
        return ctx.topics or set()

    def add_context_metadata(self, key: str, value: Any) -> None:
        ctx = logger_context.get()

        if ctx:
            ctx.context_metadata[key] = value
            logger_context.set(ctx)

    def remove_context_metadata(self, key: str) -> None:
        ctx = logger_context.get()

        if ctx:
            ctx.context_metadata.pop(key, None)
            logger_context.set(ctx)

    def add_topic(self, topic: str) -> None:
        ctx = logger_context.get()

        if ctx:
            ctx.topics.add(topic)
            logger_context.set(ctx)

    def remove_topic(self, topic: str) -> None:
        ctx = logger_context.get()

        if ctx:
            ctx.topics.discard(topic)
            logger_context.set(ctx)

    def info(self, event: Event) -> None:
        self.log_event(event, logging_level=EventLevel.INFO)

    def warning(self, event: Event) -> None:
        self.log_event(event, logging_level=EventLevel.WARNING)

    def error(self, error: HttpError) -> None:
        event_code = error.debug_details.debug_code

        if not self._should_log_based_on_sampling_percentage(event_code):
            return

        error_event = self._format_error(error)

        self.logger.error(
            error.debug_details.debug_message, extra={"event": error_event}
        )

    def log_event(self, event: Event, logging_level: EventLevel) -> None:
        if not self._should_log_based_on_sampling_percentage(event.code):
            return

        sampling_percentage = (
            self.event_codes_to_sampling_percentages_map.get(event.code, 100)
            if self.event_codes_to_sampling_percentages_map
            else 100
        )

        event_with_tracing_details = EventWithTracesDetails.from_event(
            event,
            trace_id=self.trace_id,
            level=logging_level,
            sampling_percentage=sampling_percentage,
            context_metadata=self.context_metadata,
            topics=self.topics,
            request_details=self.request_details,
        )

        extra = {"event": event_with_tracing_details}

        if logging_level == EventLevel.INFO:
            self.logger.info(event.message, extra=extra)
        elif logging_level == EventLevel.WARNING:
            self.logger.warning(event.message, extra=extra)

    def _format_error(self, error: Optional[BaseException]) -> Optional[dict[str, Any]]:
        """
        Formats the error into a string representation.
        """
        if error is None:
            return None

        if isinstance(error, HttpError):
            event_code = error.debug_details.debug_code

            original_error_details = self._format_error(error.debug_details.caused_by)

            sampling_percentage = (
                self.event_codes_to_sampling_percentages_map.get(event_code, 100)
                if self.event_codes_to_sampling_percentages_map
                else 100
            )

            error_topics = self.topics.copy()

            if error.debug_details.topic:
                error_topics.add(error.debug_details.topic)

            return ErrorEvent(
                id=str(error.debug_details.id),
                trace_id=self.trace_id,
                code=event_code,
                topics=error_topics if error_topics else None,
                level=EventLevel.ERROR,
                path=self.request_details.path if self.request_details else None,
                method=self.request_details.method if self.request_details else None,
                query_params=self.request_details.query_params if self.request_details else None,
                user_agent=self.request_details.user_agent if self.request_details else None,
                message=error.debug_details.debug_message,
                context_metadata=self.context_metadata if self.context_metadata else None,
                details=error.debug_details.debug_details,
                severity=error.debug_details.severity,
                occurred_while=error.debug_details.occurred_while,
                caused_by=original_error_details,
                status_code=error.public_details.status_code,
                public_code=error.public_details.code,
                public_message=error.public_details.message,
                public_details=error.public_details.details,
                sampling_percentage=sampling_percentage,
                timestamp=datetime.now(tz=timezone.utc).isoformat(),
            ).model_dump(mode="json")

        return {
            "type": type(error).__name__,
            "message": str(error),
            "stack_trace": traceback.format_exception(
                type(error), error, error.__traceback__
            ),
        }


    def _should_log_based_on_sampling_percentage(self, event_code: str) -> bool:
        """
        Determines whether an event should be logged based on its code and the configured sampling percentages.
        """

        if not self.event_codes_to_sampling_percentages_map:
            return True

        sampling_percentage = self.event_codes_to_sampling_percentages_map.get(event_code, 100)

        if sampling_percentage <= 0:
            return False

        if sampling_percentage >= 100:
            return True

        return random.randint(0, 100) < sampling_percentage


from .logger import Logger
from .schemas import (
    ErrorEvent,
    Event,
    EventLevel,
    EventWithTracesDetails,
    RequestDetails,
    OutgoingRequestEvent,
    IncomingRequestEvent,
    IncomingResponseEvent,
    OutgoingResponseEvent,
)
